#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import copy
import hashlib
import json
import os
import requests
import time
import traceback
import urllib
import yaml

from datetime import datetime, timedelta

from twicorder.auth import Auth
from twicorder.config import Config
from twicorder.constants import (
    DEFAULT_MONGO_OUTPUT,
    DEFAULT_OUTPUT_EXTENSION,
    TW_TIME_FORMAT
)
from twicorder.exchange import RateLimitCentral
from twicorder.project_manager import ProjectManager
from twicorder.utils import write, AppData, timestamp_to_datetime


class BaseQuery:
    """
    Base Query class.
    """
    name = NotImplemented
    endpoint = NotImplemented
    _max_count = NotImplemented
    _last_return_token = None
    _results_path = None
    _fetch_more_path = None

    _mongo_collection = None

    def __init__(self, output=None, **kwargs):
        self._done = False
        self._more_results = None
        self._results = []
        self._last_id = None
        self._output = output
        self._kwargs = kwargs
        self._orig_kwargs = copy.deepcopy(kwargs)
        self._log = []

        last_return = AppData.get_last_query_id(self.uid)
        if last_return:
            self.kwargs[self.last_return_token] = last_return

    def __eq__(self, other):
        return type(self) == type(other) and self.__dict__ == other.__dict__

    def __repr__(self):
        r = f'<Query({self.name!r}, kwargs={self.kwargs!r}) at 0x{id(self):x}>'
        return r

    def __str__(self):
        return f'{self.endpoint}\n{"-" * 80}\n{yaml.dump(self.kwargs)}'

    @property
    def config(self):
        return Config.get()

    @property
    def output(self):
        return self._output

    @property
    def last_return_token(self):
        return self._last_return_token

    @property
    def kwargs(self):
        return self._kwargs

    @property
    def max_count(self):
        return self._max_count

    @property
    def results_path(self):
        return self._results_path

    @property
    def fetch_more_path(self):
        return self._fetch_more_path

    @property
    def uid(self):
        raise NotImplementedError

    @property
    def done(self):
        return self._done

    @done.setter
    def done(self, value):
        self._done = value

    @property
    def more_results(self):
        return self._more_results

    @property
    def last_id(self):
        return self._last_id

    @last_id.setter
    def last_id(self, value):
        self._last_id = value

    @property
    def results(self):
        return self._results

    @property
    def mongo_collection(self):
        if not self.config.get('use_mongo', DEFAULT_MONGO_OUTPUT):
            return
        from twicorder import mongo
        collection = self._mongo_collection
        if not collection or not mongo.is_connected(collection):
            self._mongo_collection = mongo.create_collection()
        return self._mongo_collection

    def run(self):
        raise NotImplementedError

    def log(self, line):
        self._log.append(line)

    def fetch_log(self):
        log_data = '\n' + f' {self.endpoint} '.center(80, '=') + '\n'
        log_data += '\n'.join(self._log)
        log_data += '\n' + '=' * 80
        return log_data

    def save(self):
        if not self._results or not self._output:
            return
        save_dir = os.path.join(
            ProjectManager.output_dir,
            self._output or self.uid
        )
        extension = (
            self.config.get('save_extension') or DEFAULT_OUTPUT_EXTENSION
        )
        marker = self._results[0]
        stamp = datetime.strptime(marker['created_at'], TW_TIME_FORMAT)
        uid = marker['id']
        filename = f'{stamp:%Y-%m-%d_%H-%M-%S}_{uid}{extension}'
        file_path = os.path.join(save_dir, filename)
        results_str = '\n'.join(json.dumps(r) for r in self._results)
        write(f'{results_str}\n', file_path)
        self.log(f'Wrote {len(self.results)} tweets to "{file_path}"')

        # Write to Mongo
        if not self.mongo_collection:
            return
        try:
            for result in self._results:
                data = copy.deepcopy(result)
                data = timestamp_to_datetime(data)
                self.mongo_collection.replace_one(
                    {'id': data['id']},
                    data,
                    upsert=True
                )
        except Exception:
            self.log(f'Unable to connect to MongoDB: {traceback.format_exc()}')
        else:
            self.log(f'Wrote {len(self.results)} tweets to MongoDB')

    def bake_ids(self):
        """
        Saves a cache of tweet IDs from query result to disk. In storing the IDs
        between sessions, we make sure we don't save already found tweets.

        To prevent the disk cache growing too large, we purge IDs for tweets
        older than 14 days. Twitter's base search only goes back 7 days, so we
        shouldn't encounter tweets older than 14 days very often.
        """

        # Loading picked tweet IDs
        tweets = dict(AppData.get_query_tweets(self.name)) or {}

        # Purging tweet IDs older than 14 days
        now = datetime.now()
        old_tweets = tweets.copy()
        tweets = {}
        for tweet_id, timestamp in old_tweets.items():
            dt = datetime.fromtimestamp(timestamp)
            if not now - dt > timedelta(days=14):
                tweets[tweet_id] = timestamp

        # Stores tweet IDs from result
        self._results = [t for t in self.results if t['id'] not in tweets]
        new_tweets = []
        for result in self.results:
            created_at = result['created_at']
            dt = datetime.strptime(created_at, TW_TIME_FORMAT)
            timestamp = int(dt.timestamp())
            new_tweets.append((result['id'], timestamp))
        AppData.add_query_tweets(self.name, new_tweets)


class RequestQuery(BaseQuery):
    """
    Queries based on the requests module and the twitter API.
    """
    _base_url = 'https://api.twitter.com/1.1'
    _request_type = 'get'
    _token_auth = False

    _hash_keys = [
        'endpoint',
        '_results_path',
        '_fetch_more_path',
        '_orig_kwargs',
        '_base_url',
    ]

    def __init__(self, output=None, **kwargs):
        super(RequestQuery, self).__init__(output, **kwargs)

    def __eq__(self, other):
        return type(self) == type(other) and self.uid == other.uid

    @property
    def base_url(self):
        return self._base_url

    @property
    def request_type(self):
        return self._request_type

    @property
    def token_auth(self):
        return self._token_auth

    @property
    def request_url(self):
        url = f'{self.base_url}{self.endpoint}.json'
        if self.request_type == 'get':
            if self.kwargs:
                url += f'?{urllib.parse.urlencode(self.kwargs)}'
        return url

    @property
    def uid(self):
        hash_str = str([getattr(self, k) for k in self._hash_keys]).encode()
        return hashlib.blake2s(hash_str).hexdigest()

    def run(self):
        # Purging logs
        self._log = []

        # Check rate limit for query. Sleep if limits are in effect.
        limit = RateLimitCentral.get(self.endpoint)
        self.log(f'URL: {self.request_url}')
        self.log(f'{limit}')
        if limit and limit.remaining == 0:
            sleep_time = max(limit.reset - time.time(), 0) + 2
            msg = (
                f'Sleeping for {sleep_time:.02f} seconds for endpoint '
                f'"{self.endpoint}".'
            )
            self.log(msg)
            time.sleep(sleep_time)

        # Perform query
        attempts = 0
        while True:
            try:
                if self._token_auth:
                    request = getattr(requests, self.request_type)
                    response = request(
                        self.request_url,
                        data=json.dumps(self.kwargs),
                        auth=Auth.token()
                    )
                else:
                    request = getattr(Auth.session(), self.request_type)
                    response = request(self.request_url)
            except Exception:
                attempts += 1
                time.sleep(2**attempts)
                if attempts >= 5:
                    raise
            else:
                break

        # Check query response code. Return with error message if not a
        # successful 200 code.
        if response.status_code != 200:
            if response.status_code == 429:
                self.log(f'Rate Limit in effect: {response.reason}')
                self.log(f'Message: {response.json().get("message")}')
            else:
                self.log(
                    '<{r.status_code}> {r.reason}: {r.content}'
                    .format(r=response)
                )
            return
        self.log('Successful return!')

        # Update rate limit for query
        RateLimitCentral.update(self.endpoint, response.headers)

        # Search query response for additional paged results. Pronounce the
        # query done if no more pages are found.
        pagination = response.json()
        if self.fetch_more_path:
            for token in self.fetch_more_path.split('.'):
                pagination = pagination.get(token, {})
            if pagination:
                self._more_results = pagination
                self.log('More pages found!')
            else:
                self._more_results = None
                self._done = True
                self.log('No more pages!')
        else:
            self._done = True

        # Extract crawled tweets from query response.
        results = response.json()
        if self.results_path:
            for token in self.results_path.split('.'):
                results = results.get(token, [])
        self._results = results
        self.log(f'Result count: {len(results)}')

        # Saves and stores IDs for crawled tweets found in the query result.
        # Also records the last tweet ID found.
        if results:
            self.bake_ids()
            self.log(f'Cached Tweet IDs to disk!')
            self.save()
            if self.last_id is None:
                self.last_id = results[0].get('id_str')

        # Caches last tweet ID found to disk if the query, including all pages
        # completed successfully. This saves us from searching all the way back
        # to the beginning on next crawl. Instead we can stop when we encounter
        # this tweet.
        if self._done and self.last_id:
            self.log(f'Cached ID of last tweet returned by query to disk.')
            AppData.set_last_query_id(self.uid, self.last_id)

        # Returning crawled results
        return results



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


class TwicorderException(BaseException):
    pass


class NoConfigException(TwicorderException):

    def __init__(self, *args, **kwargs):
        if not args:
            from twicorder.project_manager import ProjectManager
            args = [(
                f'\n'
                f'No configuration file could be found for this project '
                f'({ProjectManager.project_dir}). Configs are installed to '
                f'<PROJECT_ROOT>/config/preferences.yaml.\n'
                f'\n'
                f'To set a different project directory, please run Twicorder '
                f'with the "--project-dir" option:\n'
                f'\n'
                f'twicorder --project-dir /path/to/project_dir\n'
                f'\n'
                f'See '
                f'https://github.com/thimic/twicorder-search/blob/master/'
                f'README.md for details.'
            )]
        super().__init__(*args, **kwargs)


class NoTasksException(TwicorderException):

    def __init__(self, *args, **kwargs):
        if not args:
            from twicorder.project_manager import ProjectManager
            args = [(
                f'\n'
                f'No tasks file could be found for this project '
                f'({ProjectManager.project_dir}). Tasks are installed to '
                f'<PROJECT_ROOT>/config/tasks.yaml.\n'
                f'\n'
                f'To set a different project directory, please run Twicorder '
                f'with the "--project-dir" option:\n'
                f'\n'
                f'twicorder --project-dir /path/to/project_dir\n'
                f'\n'
                f'See '
                f'https://github.com/thimic/twicorder-search/blob/master/'
                f'README.md for details.'
            )]
        super().__init__(*args, **kwargs)


class NoCredentialsException(TwicorderException):

    def __init__(self, *args, **kwargs):
        if not args:
            args = [(
                '\n'
                'No API credentials found. Credentials can be added to the '
                'project config or set as environment variables:\n'
                '\n'
                ' - TWITTER_CONSUMER_KEY\n'
                ' - TWITTER_CONSUMER_SECRET\n'
                ' - TWITTER_ACCESS_TOKEN\n'
                ' - TWITTER_ACCESS_SECRET\n'
                '\n'
                'See https://github.com/thimic/twicorder-search/blob/master/'
                'README.md for details.'
            )]
        super().__init__(*args, **kwargs)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os

from twicorder.constants import APP_DATA_TOKEN, DEFAULT_PROJECT_DIR


class _ProjectManager:
    """
    Class for handling file paths related to the current project.
    """
    _project_dir = DEFAULT_PROJECT_DIR

    @property
    def project_dir(self):
        project_path = self._project_dir
        os.makedirs(project_path, exist_ok=True)
        return project_path

    @project_dir.setter
    def project_dir(self, value):
        self._project_dir = os.path.expanduser(value)

    @property
    def config_dir(self):
        config_path = os.path.join(self.project_dir, 'config')
        os.makedirs(config_path, exist_ok=True)
        return config_path

    @property
    def output_dir(self):
        from twicorder.config import Config
        config = Config.get()
        output_path = config.get('save_dir')
        if not output_path:
            output_path = os.path.join(self.project_dir, 'tweets')
        os.makedirs(output_path, exist_ok=True)
        return output_path

    @property
    def app_data_dir(self):
        app_data_path = os.path.join(self.project_dir, 'app_data')
        os.makedirs(app_data_path, exist_ok=True)
        return app_data_path

    @property
    def logs_dir(self):
        log_path = os.path.join(self.app_data_dir, 'logs')
        os.makedirs(log_path, exist_ok=True)
        return log_path

    @property
    def preferences(self):
        return os.path.join(self.config_dir, 'preferences.yaml')

    @property
    def tasks(self):
        return os.path.join(self.config_dir, 'tasks.yaml')
    
    @property
    def logs(self):
        return os.path.join(self.logs_dir, f'{APP_DATA_TOKEN}.log')

    @property
    def app_data(self):
        return os.path.join(self.app_data_dir, f'{APP_DATA_TOKEN}.sql')


ProjectManager = _ProjectManager()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import urllib

from twicorder.cached_users import CachedUserCentral
from twicorder.constants import DEFAULT_EXPAND_USERS
from twicorder.queries import RequestQuery


class StatusQuery(RequestQuery):

    name = 'status'
    endpoint = '/statuses/lookup'

    def __init__(self, output=None, **kwargs):
        super(StatusQuery, self).__init__(output, **kwargs)
        self._kwargs['tweet_mode'] = 'extended'
        self._kwargs['include_entities'] = 'true'
        self._kwargs['trim_user'] = 'false'
        self._kwargs.update(kwargs)

    def save(self):
        for status in self.results:
            pass
        msg = f'Save for endpoint "{self.endpoint}" is not yet implemented.'
        raise NotImplementedError(msg)


class TimelineQuery(RequestQuery):

    name = 'user_timeline'
    endpoint = '/statuses/user_timeline'
    _last_return_token = 'since_id'

    def __init__(self, output=None, **kwargs):
        super(TimelineQuery, self).__init__(output, **kwargs)
        self._kwargs['tweet_mode'] = 'extended'
        self._kwargs['result_type'] = 'recent'
        self._kwargs['count'] = 200
        self._kwargs['trim_user'] = 'false'
        self._kwargs['exclude_replies'] = 'false'
        self._kwargs['include_rts'] = 'true'
        self._kwargs.update(kwargs)

    @property
    def request_url(self):
        url = f'{self.base_url}{self.endpoint}.json'
        if self.request_type == 'get':
            if self.more_results:
                self.kwargs['max_id'] = self.more_results
            url += f'?{urllib.parse.urlencode(self.kwargs)}'
        return url

    def run(self):
        super(TimelineQuery, self).run()
        self.done = False
        if not self.results:
            self.done = True
            return
        self._more_results = self.results[-1]['id_str']
        last_return = self.kwargs.get('max_id')
        if last_return and int(self._more_results) >= int(last_return):
            self.done = True

    def save(self):
        if self.config.get('full_user_mentions', DEFAULT_EXPAND_USERS):
            self.log('Expanding user mentions!')
            CachedUserCentral.expand_user_mentions(self.results)
        super(TimelineQuery, self).save()


class StandardSearchQuery(RequestQuery):

    name = 'free_search'
    endpoint = '/search/tweets'
    _last_return_token = 'since_id'
    _results_path = 'statuses'
    _fetch_more_path = 'search_metadata.next_results'

    def __init__(self, output=None, **kwargs):
        super(StandardSearchQuery, self).__init__(output, **kwargs)
        self._kwargs['tweet_mode'] = 'extended'
        self._kwargs['result_type'] = 'recent'
        self._kwargs['count'] = 100
        self._kwargs['include_entities'] = 'true'
        self._kwargs.update(kwargs)

    @property
    def request_url(self):
        url = f'{self.base_url}{self.endpoint}.json'
        if self.request_type == 'get':
            if self.more_results:
                url += self.more_results
                # API bug: 'search_metadata.next_results' does not include
                # 'tweet_mode'. Adding it back in manually.
                if 'tweet_mode=extended' not in url:
                    url += '&tweet_mode=extended'
            elif self.kwargs:
                url += f'?{urllib.parse.urlencode(self.kwargs)}'
        return url

    def save(self):
        if self.config.get('full_user_mentions', DEFAULT_EXPAND_USERS):
            self.log('Expanding user mentions!')
            CachedUserCentral.expand_user_mentions(self.results)
        super(StandardSearchQuery, self).save()


class FullArchiveGetQuery(RequestQuery):

    name = 'fullarchive_get'
    endpoint = '/tweets/search/fullarchive/production'
    _fetch_more_path = 'next'


class FullArchivePostQuery(RequestQuery):

    name = 'fullarchive_post'
    endpoint = '/tweets/search/fullarchive/production'
    _fetch_more_path = 'next'
    _request_type = 'post'
    _token_auth = True


class FriendsList(RequestQuery):

    name = 'friends_list'
    endpoint = '/friends/list'


class RateLimitStatusQuery(RequestQuery):

    name = 'rate_limit_status'
    endpoint = '/application/rate_limit_status'




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import time

from datetime import datetime
from queue import Queue
from threading import Thread

from twicorder.utils import TwiLogger

logger = TwiLogger()


class RateLimitCentral:
    """
    Class keeping track of end points and their rate limits.
    """
    _limits = {}

    @classmethod
    def update(cls, endpoint, header):
        """
        Update endpoint with latest rate limit information.

        Args:
            endpoint (str): Endpoint
            header (dict): Query response header

        """
        limit_keys = {
            'x-rate-limit-limit',
            'x-rate-limit-remaining',
            'x-rate-limit-reset'
        }
        if not limit_keys.issubset(header.keys()):
            return
        cls._limits[endpoint] = RateLimit(header)

    @classmethod
    def get(cls, endpoint):
        """
        Retrieves latest rate limit information for the given endpoint.
        Args:
            endpoint (str): Endpoint

        Returns:
            RateLimit: Rate limit object

        """
        return cls._limits.get(endpoint)

    @classmethod
    def get_cap(cls, endpoint):
        """
        Retrieve the query cap for the given endpoint.

        Args:
            endpoint (str): Endpoint

        Returns:
            int: Max queries per 15 minutes

        """
        limit = cls.get(endpoint)
        if not limit:
            return
        return limit.cap

    @classmethod
    def get_remaining(cls, endpoint):
        """
        Retrieve number of remaining queries for the given endpoint.

        Args:
            endpoint (str): Endpoint

        Returns:
            int: Remaining queries for the current 15 minute window

        """
        limit = cls.get(endpoint)
        if not limit:
            return
        return limit.remaining

    @classmethod
    def get_reset(cls, endpoint):
        """
        Retrieve time until the current 15 minute window expires.

        Args:
            endpoint (str): Endpoint

        Returns:
            float: Time in seconds

        """
        limit = cls.get(endpoint)
        if not limit:
            return
        return limit.reset


class RateLimit:
    """
    Rate limit object, used to describe the limits for a given API end point.
    """
    def __init__(self, headers):
        self._cap = headers.get('x-rate-limit-limit')
        self._remaining = int(headers.get('x-rate-limit-remaining'))
        self._reset = float(headers.get('x-rate-limit-reset'))

    def __repr__(self):
        reset = datetime.fromtimestamp(self._reset)
        representation = (
            f'RateLimit(cap={self.cap}, remaining={self.remaining}, '
            f'reset="{reset:%y.%m.%d %H:%M:%S}")'
        )
        return representation

    @property
    def cap(self):
        """
        Queries allowed per 15 minutes.

        Returns:
            int: Number of queries

        """
        return self._cap

    @property
    def remaining(self):
        """
        Queries left for the current 15 minute window.

        Returns:
            int: Number of queries

        """
        return self._remaining

    @property
    def reset(self):
        """
        Time until the current 15 minute window expires.

        Returns:
            float: Reset time

        """
        return self._reset


class QueryWorker(Thread):
    """
    Queue thread, used to execute queue queries.
    """

    def __init__(self, *args, **kwargs):
        super(QueryWorker, self).__init__(*args, **kwargs)
        self._query = None

    def setup(self, queue):
        self._queue = queue

    @property
    def queue(self):
        return self._queue

    @property
    def query(self):
        return self._query

    def run(self):
        """
        Fetches query from queue and executes it.
        """
        while True:
            self._query = self.queue.get()
            if self.query is None:
                logger.info(f'Terminating thread "{self.name}"')
                break
            while not self.query.done:
                try:
                    self.query.run()
                except Exception:
                    logger.exception('Query failed:\n')
                    break
                logger.info(self.query.fetch_log())
                time.sleep(.2)
            time.sleep(.5)
            self.queue.task_done()


class QueryExchange:
    """
    Organises queries in queues and executes them after the FIFO principle.
    """

    queues = {}
    threads = {}
    failure = False

    @classmethod
    def get_queue(cls, endpoint):
        """
        Retrieves the queue for the given endpoint if it exists, otherwise
        creates a queue.

        Args:
            endpoint (str): API endpoint

        Returns:
            Queue: Queue for endpoint

        """
        if not cls.queues.get(endpoint):
            queue = Queue()
            cls.queues[endpoint] = queue
            thread = QueryWorker(name=endpoint)
            thread.setup(queue=queue)
            thread.start()
            cls.threads[endpoint] = thread
        return cls.queues[endpoint]

    @classmethod
    def add(cls, query):
        """
        Finds appropriate queue for given end point and adds it.

        Args:
            query (BaseQuery): Query object

        """
        queue = cls.get_queue(query.endpoint)
        if query in queue.queue:
            logger.info(f'Query with ID {query.uid} is already in the queue.')
            return
        thread = cls.threads.get(query.endpoint)
        if thread and thread.query == query:
            logger.info(f'Query with ID {query.uid} is already running.')
            return
        queue.put(query)

    @classmethod
    def clear(cls):
        """
        Prepares QueryExchange for a new run.
        """
        cls.queues = {}
        cls.threads = {}
        cls.failure = False

    @classmethod
    def wait(cls):
        """
        Sends shutdown signal to threads and waits for all threads and queues to
        terminate.
        """
        for queue in cls.queues.values():
            queue.put_nowait(None)
        for thread in cls.threads.values():
            thread.join()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import glob
import json
import os

from datetime import datetime

from pymongo import MongoClient, TEXT
from pymongo.collection import Collection
from pymongo.database import Database
from pymongo.errors import ServerSelectionTimeoutError

from twicorder import utils
from twicorder.config import Config
from twicorder.project_manager import ProjectManager
from twicorder.utils import TwiLogger

logger = TwiLogger()


def is_connected(entity):
    """
    Checks if client is still connected to database.

    Args:
        entity (object): MongoClient, Database or Collection object

    Returns:
        bool: True if connected, else False

    """
    if isinstance(entity, MongoClient):
        client = entity
    elif isinstance(entity, Database):
        client = entity.client
    elif isinstance(entity, Collection):
        client = entity.database.client
    else:
        error = (
            'Please provide an instance of MongoClient, Database or Collection.'
        )
        raise TypeError(error)
    try:
        client.server_info()
    except ServerSelectionTimeoutError:
        return False
    return True


def create_collection(db_name='twicorder', collection_name='tweets'):
    """
    Create collection for the given database. Skip an return early if collection
    exists.

    Args:
        db_name (str): Database name
        collection_name (str): Collection name

    Returns:
        Collection: Created collection.

    """
    try:
        client = MongoClient()
        if not is_connected(client):
            return
        db = client[db_name]
        if collection_name in db.list_collection_names():
            return db[collection_name]
        collection = db[collection_name]
        collection.create_index('id', unique=True)
        collection.create_index('created_at')
        collection.create_index('retweet_count')
        collection.create_index('favorite_count')
        collection.create_index('in_reply_to_status_id')
        collection.create_index('in_reply_to_user_id')
        collection.create_index('in_reply_to_screen_name')
        collection.create_index('entities.hashtags')
        collection.create_index('user.created_at')
        collection.create_index('user.screen_name')
        collection.create_index('user.id')
        collection.create_index('user.followers_count')
        collection.create_index('user.favourites_count')
        collection.create_index('user.verified')
        collection.create_index('user.statuses_count')
        collection.create_index(
            [('full_text', TEXT)],
            default_language='english'
        )
        return collection
    except Exception:
        logger.exception('Unable to connect to MongoDB: ')
        return


def backfill(path=None, db_name='twicorder', collection_name='tweets'):
    logger = TwiLogger()
    tweets = create_collection(db_name, collection_name)

    save_dir = os.path.expanduser(path or ProjectManager.output_dir)

    paths = glob.glob(os.path.join(save_dir, '**', '*.t*'), recursive=True)
    t0 = datetime.now()
    for idx, path in enumerate(paths):
        if os.path.basename(os.path.dirname(path)) != 'stream':
            continue
        try:
            for lidx, line in enumerate(utils.readlines(path)):
                try:
                    data = json.loads(line)
                except Exception:
                    logger.exception(
                        f'Backfill: Unable to read line {path}:{lidx + 1}'
                    )
                    continue
                else:
                    if data.get('delete'):
                        continue
                    if os.path.basename(os.path.dirname(path)) == 'stream':
                        data = utils.stream_to_search(data)
                    data = utils.timestamp_to_datetime(data)
                    tweets.replace_one({'id': data['id']}, data, upsert=True)
            t_delta = datetime.now() - t0
            average = t_delta / (idx + 1)
            remaining = str((len(paths) - (idx + 1)) * average).split('.')[0]

            logger.info(
                f'{idx + 1}/{len(paths)} '
                f'{remaining} '
                f'{os.sep.join(path.split(os.sep)[-3:])}'
            )
        except Exception:
            logger.exception(f'Backfill: Unable to read file: {path}')


if __name__ == '__main__':
    backfill()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import inspect
import sys
import time

from threading import Thread

from twicorder import TwicorderException
from twicorder.project_manager import ProjectManager


class WorkerThread(Thread):
    """
    Background thread, running queries.
    """
    def setup(self, func, tasks):
        """
        Preparing thread.

        Args:
            func (func): Function to perform
            tasks (list[twicorder.tasks.Task]): Tasks to perform

        """
        self._running = False
        self._func = func
        self._tasks = tasks

    def stop(self):
        """
        Stop thread.
        """
        self._running = False

    def run(self):
        """
        Start thread.
        """
        from twicorder.exchange import QueryExchange
        self._running = True
        logger.info(' Loading tasks '.center(80, '='))
        logger.info('')
        while self._running:
            update = False
            for task in self._tasks:
                if not task.due:
                    continue
                update = True
                query = self._func(task)
                QueryExchange.add(query)
                logger.info(query)
            if update:
                logger.info('=' * 80)
            # Sleep 1 minute, then wake up and check if any queries are due to
            # run.
            time.sleep(60)


class Twicorder:
    """
    Twicorder controller class.
    """
    def __init__(self, project_dir=None):
        """
        Constructor for Twicorder class. Sets up the task manager, query
        exchange, worker thread and query types.

        Keyword Args:
            project_dir (str): Path to Twicorder project directory

        """
        if project_dir:
            ProjectManager.project_dir = project_dir

        # Todo: Only import logger after project dir is set, to ensure logging
        #       to project dir. This is ugly and needs a better solution.
        from twicorder.utils import TwiLogger
        global logger
        logger = TwiLogger()

        # Test setup before continuing
        try:
            from twicorder.config import Config
            from twicorder.tasks import TaskManager
            from twicorder.auth import Auth
            Config.get()
            TaskManager.load()
            Auth.session()
        except TwicorderException as error:
            logger.critical(error)
            sys.exit(1)
            return

        from twicorder.tasks import TaskManager
        self._task_manager = TaskManager()
        self._worker_thread = WorkerThread()
        self._query_types = {}

    @property
    def task_manager(self):
        """
        Task manager instance for Twicorder.

        Returns:
            twicorder.tasks.TaskManager: Task manager instance

        """
        return self._task_manager

    @property
    def tasks(self):
        """
        List of tasks registered in the task manager.

        Returns:
            list[twicorder.tasks.Task]: List of tasks

        """
        return self._task_manager.tasks

    @property
    def query_types(self):
        """
        Compiles a dictionary of available query types by inspecting the
        twicorder.queries.request_queries module.

        Returns:
            dict: Available query types by name

        """
        if self._query_types:
            return self._query_types
        from twicorder.queries import RequestQuery
        from twicorder.queries import request_queries
        for name, item in inspect.getmembers(request_queries, inspect.isclass):
            if item == RequestQuery:
                continue
            elif issubclass(item, RequestQuery):
                self._query_types[item.name] = item
        return self._query_types

    def stop(self):
        """
        Stops crawler.
        """
        self._worker_thread.stop()

    def run(self):
        """
        Starts crawler.
        """
        self._worker_thread.setup(
            func=self.cast_query,
            tasks=self.tasks,
        )
        self._worker_thread.start()

    def cast_query(self, task):
        """
        Casts the given task to a query.

        Args:
            task (twicorder.tasks.Task): Task

        Returns:
            twicorder.queries.BaseQuery: Query


        """
        query_object = self.query_types[task.name]
        query = query_object(task.output, **task.kwargs)
        return query


if __name__ == '__main__':
    twicorder = Twicorder('~/Desktop/Twicorder')
    twicorder.run()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import logging
import os
import sqlite3
import sys

from datetime import datetime
from gzip import GzipFile
from logging import StreamHandler
from threading import Lock

from logging.handlers import RotatingFileHandler

from twicorder.config import Config
from twicorder.constants import (
    COMPRESSED_EXTENSIONS,
    DEFAULT_APP_DATA_CONNECTION_TIMEOUT,
    REGULAR_EXTENSIONS,
    TW_TIME_FORMAT,
)
from twicorder.project_manager import ProjectManager


class TwiLogger:

    _logger = None

    @classmethod
    def setup(cls):
        cls._logger = logging.getLogger('Twicorder')
        file_handler = RotatingFileHandler(
            ProjectManager.logs,
            maxBytes=1024**2 * 10,
            backupCount=5
        )
        formatter = logging.Formatter(
            '%(asctime)s: [%(levelname)s] %(message)s'
        )
        file_handler.setFormatter(formatter)
        file_handler.setLevel(logging.WARNING)
        cls._logger.addHandler(file_handler)

        stream_handler = StreamHandler(sys.stdout)
        stream_handler.setLevel(logging.DEBUG)
        cls._logger.addHandler(stream_handler)

        cls._logger.setLevel(logging.DEBUG)

    def __new__(cls, *args, **kwargs):
        if not cls._logger:
            cls.setup()
        return cls._logger


def auto_commit(func):
    def func_wrapper(self, *args, **kwargs):
        with self._conn:
            func(self, *args, **kwargs)
    return func_wrapper


class AppData:
    """
    Class for reading and writing AppData to be used between sessions.
    """

    _config = Config.get()
    _timeout = (
        _config.get('appdata_connection_timeout') or
        DEFAULT_APP_DATA_CONNECTION_TIMEOUT
    )
    _con = sqlite3.connect(
        ProjectManager.app_data,
        check_same_thread=False,
        timeout=float(_timeout)
    )
    _lock = Lock()

    def __del__(self):
        self._con.close()

    @classmethod
    def _make_query_table(cls, name):
        with cls._lock, cls._con as con:
            con.execute(
                f'''
                CREATE TABLE IF NOT EXISTS [{name}] (
                    tweet_id INTEGER PRIMARY KEY,
                    timestamp INTEGER NOT NULL
                )
                '''
            )

    @classmethod
    def _make_last_id_table(cls):
        with cls._lock, cls._con as con:
            con.execute(
                '''
                CREATE TABLE IF NOT EXISTS queries_last_id (
                    query_hash TEXT PRIMARY KEY,
                    tweet_id INTEGER NOT NULL
                )
                '''
            )

    @classmethod
    def add_query_tweet(cls, query_name, tweet_id, timestamp):
        cls._make_query_table(query_name)
        with cls._lock, cls._con as con:
            con.execute(
                f'''
                INSERT OR REPLACE INTO {query_name} VALUES (
                    ?, ?
                )
                ''',
                (tweet_id, timestamp)
            )

    @classmethod
    def add_query_tweets(cls, query_name, tweets):
        cls._make_query_table(query_name)
        with cls._lock, cls._con as con:
            con.executemany(
                f'''
                INSERT OR REPLACE INTO {query_name} VALUES (
                    ?, ?
                )
                ''',
                tweets
            )

    @classmethod
    def get_query_tweets(cls, query_name):
        cls._make_query_table(query_name)
        with cls._lock, cls._con as con:
            cursor = con.cursor()
            cursor.execute(
                f'''
                SELECT DISTINCT
                    tweet_id, timestamp
                FROM
                    {query_name}
                '''
            )
            return cursor.fetchall()

    @classmethod
    def set_last_query_id(cls, query_hash, tweet_id):
        cls._make_last_id_table()
        with cls._lock, cls._con as con:
            con.execute(
                '''
                INSERT OR REPLACE INTO queries_last_id VALUES (
                    ?, ?
                )
                ''',
                (query_hash, tweet_id)
            )

    @classmethod
    def get_last_query_id(cls, query_hash):
        cls._make_last_id_table()
        with cls._lock, cls._con as con:
            cursor = con.cursor()
            cursor.execute(
                '''
                SELECT
                DISTINCT
                    tweet_id
                FROM
                    queries_last_id
                WHERE
                    query_hash=?
                ''',
                (query_hash,)
            )
            result = cursor.fetchone()
        if not result:
            return
        return result[0]


def twopen(filename, mode='r'):
    """
    Replacement method for Python's build-in open. Adds the option to handle
    compressed files.

    Args:
        filename (str): Path to file
        mode (str): Open mode

    Returns:
        TextIOWrapper / GzipFile: File object

    Raises:
        IOError: If extension is unknown.

    """
    filename = os.path.expanduser(filename)
    dirname = os.path.dirname(filename)
    if mode in ('a', 'w') and not os.path.isdir(dirname):
        os.makedirs(dirname)
    ext = os.path.splitext(filename)[-1].strip('.')
    if ext in REGULAR_EXTENSIONS:
        return open(file=filename, mode=mode)
    elif ext in COMPRESSED_EXTENSIONS:
        return GzipFile(filename=filename, mode=mode)
    else:
        raise IOError('Unrecognised format: {}'.format(ext))


def read(filename):
    """
    Reading the file for a given path.

    Args:
        filename (str): Path to file to read

    Returns:
        str: File data

    """
    with twopen(filename=filename, mode='r') as file_object:
        data = file_object.read()
        if isinstance(file_object, GzipFile):
            data = data.decode('utf-8')
        return data


def readlines(filename):
    """
    Reading the file for a given path.

    Args:
        filename (str): Path to file to read

    Returns:
        str: File data

    """
    with twopen(filename=filename, mode='r') as file_object:
        data = file_object.readlines()
        if isinstance(file_object, GzipFile):
            data = [d.decode('utf-8') for d in data]
        return data


def write(data, filename, mode='a'):
    """
    Appending data to the given file.

    Args:
        data (str): Data to append to the given file
        filename (str): Path to file to write
        mode (str): File stream mode ('a'. 'w' etc)

    """
    with twopen(filename=filename, mode=mode) as file_object:
        if isinstance(file_object, GzipFile):
            file_object.write(data.encode('utf-8'))
            return
        file_object.write(data)


def message(title='Warning', body='', width=80):
    """
    Prints a formatted message based on input

    Args:
        title (str): Title of the message
        body (str): Message body
        width (int): Message line width

    """
    header = ' {} '.format(title).center(width, '=')
    footer = '=' * width
    text = (
        '\n'
        '{}\n'
        '\n'
        '{}\n'
        '\n'
        '{}\n'
        '\n'
    )
    print(text.format(header, body, footer))


def collect_key_values(key, data):
    """
    Builds a list of values for all keys matching the given "key" in a nested
    dictionary.

    Args:
        key (object): Dictionary key to search for
        data (dict): Nested data dict

    Returns:
        list: List of values for given key

    """
    values = []
    for k, v in data.items():
        if k == key:
            values.append(v)
            continue
        if isinstance(v, dict):
            values += collect_key_values(key, v)
    return values


def flatten(l):
    """
    Flattens a nested list

    Args:
        l (list): Nested list

    Returns:
        list: Flattened list

    """
    return [item for sublist in l for item in sublist]


def str_to_date(text):
    """
    Turns a time stamp represented as a string into a datetime object.

    Args:
        text (str): Time stamp

    Returns:
        datetime.datetime: Time stamp as datetime object

    """
    return datetime.strptime(text, TW_TIME_FORMAT)


def timestamp_to_datetime(data):
    """
    Traverse dictionary and convert all instances of time stamp strings into
    datetime objects.

    Args:
        data (dict): Tweet dictionary

    Returns:
        dict: Updated tweet dictionary

    """
    for key, value in data.items():
        if key in ['created_at', 'recorded_at'] and isinstance(value, str):
            data[key] = datetime.strptime(value, TW_TIME_FORMAT)
        elif isinstance(value, dict):
            data[key] = timestamp_to_datetime(value)
        elif isinstance(value, list):
            data[key] = [
                timestamp_to_datetime(v) for v in value if isinstance(v, dict)
            ]
    return data


def stream_to_search(data):
    """
    Conform tweet dictionaries collected from the streaming API to the format of
    tweets collected from the search API.

    Args:
        data (dict): Tweet dictionary

    Returns:
        dict: Updated tweet dictionary

    """
    extended_tweet = data.get('extended_tweet')
    if extended_tweet:
        data.pop('extended_tweet')
        data.update(extended_tweet)
        data['truncated'] = False
        data.pop('text')
    else:
        if data.get('text'):
            data['full_text'] = data.pop('text')
    for key, value in data.items():
        if key in ['retweeted_status', 'quoted_status']:
            data[key] = stream_to_search(value)
    return data



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import time
import yaml

from twicorder import NoTasksException
from twicorder.project_manager import ProjectManager
from twicorder.utils import TwiLogger

logger = TwiLogger()


class Task:

    def __init__(self, name, frequency=15, output=None, **kwargs):
        self._name = name
        self._frequency = frequency
        self._output = output
        self._kwargs = kwargs

        self._last_run = None

    def __eq__(self, other):
        return type(self) == type(other) and self.__dict__ == other.__dict__

    def __repr__(self):
        string = (
            f'Task('
            f'name={repr(self.name)}, '
            f'frequency={repr(self.frequency)}, '
            f'multipart={self.multipart}, '
            f'kwargs={str(self.kwargs)}'
            f')'
        )
        return string

    @property
    def name(self):
        return self._name

    @property
    def frequency(self):
        return self._frequency

    @property
    def output(self):
        return self._output

    @property
    def kwargs(self):
        return self._kwargs

    @property
    def due(self):
        if self._last_run is None:
            self._last_run = time.time()
            return True
        if time.time() - self._last_run >= self.frequency * 60:
            self._last_run = time.time()
            return True
        return False


class TaskManager:

    _tasks = []

    @classmethod
    def load(cls):
        """
        Reading tasks from yaml file and parsing to a dictionary.
        """
        cls._tasks = []
        if not os.path.isfile(ProjectManager.tasks):
            raise NoTasksException
        with open(ProjectManager.tasks, 'r') as stream:
            raw_tasks = yaml.full_load(stream)
        for query, tasks in raw_tasks.items():
            for raw_task in tasks or []:
                task = Task(
                    name=query,
                    frequency=raw_task.get('frequency') or 15,
                    output=raw_task.get('output'),
                    **raw_task.get('kwargs') or {}
                )
                cls._tasks.append(task)

    @property
    def tasks(self):
        if not self._tasks:
            self.load()
        return self._tasks



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import requests

from requests.auth import AuthBase
from requests_oauthlib import OAuth1Session

from twicorder import NoCredentialsException
from twicorder.config import Config
from twicorder.utils import TwiLogger


logger = TwiLogger()


class Auth:
    """
    Class for handling API authentication.
    """

    _config = Config.get()
    _consumer_key = (
        _config.get('consumer_key') or os.getenv('TWITTER_CONSUMER_KEY')
    )
    _consumer_secret = (
        _config.get('consumer_secret') or os.getenv('TWITTER_CONSUMER_SECRET')
    )
    _access_token = (
        _config.get('access_token') or os.getenv('TWITTER_ACCESS_TOKEN')
    )
    _access_secret = (
        _config.get('access_secret') or os.getenv('TWITTER_ACCESS_SECRET')
    )

    _session = None
    _bearer_token = None

    @classmethod
    def session(cls):
        """
        Creates an OAuth1Session or returns an existing one.

        Returns:
            OAuth1Session: Logged in session

        """
        if not (cls._consumer_key and cls._consumer_secret):
            raise NoCredentialsException
        if not cls._session:
            cls._session = OAuth1Session(
                client_key=cls._consumer_key,
                client_secret=cls._consumer_secret
            )
        return cls._session

    @classmethod
    def token(cls):
        """
        Requests an OAuth2 bearer token from the API.

        Returns:
            OAuth2Bearer: Token object

        """
        if not (cls._consumer_key and cls._consumer_secret):
            raise NoCredentialsException
        if not cls._bearer_token:
            resp = requests.post(
                'https://api.twitter.com/oauth2/token',
                auth=(os.getenv('CONSUMER_KEY'), os.getenv('CONSUMER_SECRET')),
                data={'grant_type': 'client_credentials'}
            )
            data = resp.json()
            token_type = data.get('token_type')
            if token_type != 'bearer':
                msg = (
                    f'Expected token_type to equal "bearer", but got '
                    f'{token_type} instead.'
                )
                raise AttributeError(msg)

            cls._bearer_token = OAuth2Bearer(data['access_token'])
        return cls._bearer_token


class OAuth2Bearer(AuthBase):
    """
    Bearer token implementation for requests.auth.
    """
    def __init__(self, bearer_token):
        self.bearer_token = bearer_token

    def __call__(self, request):
        request.headers['Authorization'] = 'Bearer ' + self.bearer_token
        return request



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from datetime import datetime, timedelta
from threading import Lock

from twicorder.utils import collect_key_values
from twicorder.config import Config
from twicorder.constants import DEFAULT_EXPAND_USERS_INTERVAL
from twicorder.queries import RequestQuery


class UserQuery(RequestQuery):

    name = 'user'
    endpoint = '/users/lookup'

    def __init__(self, output=None, **kwargs):
        super(UserQuery, self).__init__(output, **kwargs)
        self._kwargs['tweet_mode'] = 'extended'
        self._kwargs['include_entities'] = 'true'
        self._kwargs.update(kwargs)

    def bake_ids(self):
        return

    def save(self):
        for user in self.results:
            CachedUserCentral.add(user)


class CachedUser:
    """
    Class storing information about one cached user.
    """
    def __init__(self, user_data):
        """
        CachedUser constructor. Takes raw user data and records time of
        creation.

        Args:
            user_data (dict): Raw user data

        """

        self._data = user_data
        self._timestamp = datetime.now()

    def __eq__(self, other):
        return type(self) == type(other) and self.uid == other.uid

    def __repr__(self):
        rep = (
            f'CachedUser('
            f'name={self.screen_name}, '
            f'timestamp={self.timestamp:%c}'
            f')'
        )
        return rep

    @property
    def uid(self):
        """
        User ID.

        Returns:
            int: Unique user ID

        """
        return self._data['id']

    @property
    def screen_name(self):
        """
        User screen name.

        Returns:
            str: User screen name

        """
        return self._data['screen_name']

    @property
    def data(self):
        """
        Raw data for user.

        Returns:
            dict: Raw user data

        """
        return self._data

    @property
    def timestamp(self):
        """
        User capture time.

        Returns:
            datetime.datetime: Time stamp for user capture

        """
        return self._timestamp


class CachedUserCentral:
    """
    A class to cache user lookups for a certain amount of time. This saves
    performing the same user lookup too frequently, which can lead to hitting
    rate limit caps.
    """

    users = {}
    _config = Config.get()
    _cache_life = timedelta(
        minutes=_config.get(
            'user_lookup_interval',
            DEFAULT_EXPAND_USERS_INTERVAL
        )
    )
    lock = Lock()

    @classmethod
    def add(cls, user):
        """
        Adds the given user to the cache.

        Args:
            user (dict): Raw user data

        """
        cls.users[user['id']] = CachedUser(user)

    @classmethod
    def filter(cls):
        """
        Filters out expired users from cache.
        """
        cls.users = {
            k: v for k, v in cls.users.items()
            if datetime.now() - v.timestamp <= cls._cache_life
        }

    @classmethod
    def expand_user_mentions(cls, tweets):
        """
        Expands user mentions for tweets in result. Performs API user lookup if
        no data is found for the given mention.

        Args:
            tweets (list[dict]): List of tweets

        Returns:
            list[dict]: List of tweets with expanded user mentions

        """
        with cls.lock:
            cls.filter()
            missing_users = set([])
            for tweet in tweets:
                for user in collect_key_values('user', tweet):
                    cls.add(user)
                mention_sections = collect_key_values('user_mentions', tweet)
                for mention_section in mention_sections:
                    for mention in mention_section:
                        if not mention['id'] in cls.users:
                            missing_users.add(mention['id'])
            if not missing_users:
                return
            missing_users = list(missing_users)
            n = 100
            chunks = [
                missing_users[i:i + n] for i in range(0, len(missing_users), n)
            ]
            for chunk in chunks:
                UserQuery(user_id=','.join([str(u) for u in chunk])).run()
            for tweet in tweets:
                mention_sections = collect_key_values('user_mentions', tweet)
                for mention_section in mention_sections:
                    for mention in mention_section:
                        full_user = cls.users.get(mention['id'])
                        if not full_user:
                            continue
                        mention.update(full_user.data)
        return tweets



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os

DEFAULT_PROJECT_DIR = os.path.join(os.path.expanduser('~'), 'Twicorder')
DEFAULT_OUTPUT_EXTENSION = '.zip'

DEFAULT_CONFIG_RELOAD_INTERVAL = 15
DEFAULT_MONGO_OUTPUT = False
DEFAULT_EXPAND_USERS = False
DEFAULT_EXPAND_USERS_INTERVAL = 15

TW_TIME_FORMAT = '%a %b %d %H:%M:%S %z %Y'

REGULAR_EXTENSIONS = ['txt', 'json', 'yaml', 'twc']
COMPRESSED_EXTENSIONS = ['gzip', 'zip', 'twzip']

COMPANY = 'Zhenyael'
APP = 'Twicorder'

APP_DATA_TOKEN = 'twicorder'
DEFAULT_APP_DATA_CONNECTION_TIMEOUT = 5.0



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from distutils.core import setup


setup(
    name='twicorder-search',
    packages=['twicorder', 'twicorder/queries'],
    scripts=['bin/twicorder', 'bin/twiread'],
    version='0.2.9',
    license='MIT',
    description='A Twitter crawler for Python 3 based on Twitter\'s public API',
    author='Michael Thingnes',
    author_email='thimic@gmail.com',
    url='https://github.com/thimic/twicorder-search',
    download_url=(
        'https://github.com/thimic/twicorder-search/archive/v0.2.9.tar.gz'
    ),
    keywords=['TWITTER', 'CRAWLER', 'RESEARCH'],
    install_requires=[
        'click',
        'pymongo',
        'pyyaml',
        'requests',
        'requests_oauthlib',
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from twicorder.exceptions import (
    NoConfigException,
    NoCredentialsException,
    NoTasksException,
    TwicorderException
)
from twicorder.controller import Twicorder


__version__ = '0.2.9'



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import datetime
import os
import yaml

from twicorder import NoConfigException
from twicorder.constants import DEFAULT_CONFIG_RELOAD_INTERVAL
from twicorder.project_manager import ProjectManager


class Config:
    """
    Class for reading config file. Re-checking file on disk after a set
    interval to pick up changes.
    """

    _cache = None
    _cache_time = None
    _last_path = None

    @classmethod
    def _load(cls):
        """
        Reading config file from disk and parsing to a dictionary using the
        yaml module.

        Returns:
            dict: Config object

        """
        if not os.path.isfile(ProjectManager.preferences):
            raise NoConfigException
        with open(ProjectManager.preferences, 'r') as stream:
            config = yaml.full_load(stream)
        return config

    @classmethod
    def get(cls):
        """
        Reads config file from disk if no config object has been loaded or if
        the available config object has expired. Otherwise serving up a cached
        config object.

        Returns:
            dict: Config object

        """
        if not cls._cache or ProjectManager.preferences != cls._last_path:
            cls._cache = cls._load()
            cls._cache_time = datetime.datetime.now()
            return cls._cache
        reload_interval = (
            cls._cache['config_reload_interval'] or
            DEFAULT_CONFIG_RELOAD_INTERVAL
        )
        max_interval = datetime.timedelta(seconds=reload_interval)
        if datetime.datetime.now() - cls._cache_time > max_interval:
            cls._cache = cls._load()
            cls._cache_time = datetime.datetime.now()
        return cls._cache
